"""
This module handles the fetching of additional data like wiki pages.
"""
import time
import datetime

import requests
from sqlalchemy import select, func
from sqlalchemy.orm import undefer
import tqdm

from .db.models import Subreddit, Post, WikiPage, SubredditRule
from .importer import import_posts, import_comments
from .util import get_urls_from_string, parse_reddit_url


def reddit_reference_to_url(reference, to_root):
    """
    Generate a ZIM-internal URL for the reference.

    @param reference: reference to turn into a URL
    @type reference: L{dict}
    @param to_root: rootification prefix, just like in templates
    @type to_root: L{str}
    @return: the generated URL
    @rtype: L{str}
    """
    if reference["type"] == "subreddit":
        postfix = "/r/{}/".format(reference["subreddit"])
    elif reference["type"] == "post":
        postfix = "/r/{}/{}/".format(reference["subreddit"], reference["post"])
    elif reference["type"] == "comment":
        postfix = "/r/{}/{}/#comment-{}".format(reference["subreddit"], reference["post"], reference["comment"])
    else:
        raise ValueError("Unknown reference: {}".format(reference))
    return "//" + to_root + postfix


def get_reddit_references_from_post(post):
    """
    Return all reddit references contained in a post.

    @param post: post to get references from
    @type post: L{arcticzim.db.models.Post}
    @return: all references in the post, like L{parse_reddit_url}
    @rtype: L{list} of L{dict}
    """
    # get all URLs
    urls = [post.url]
    urls += get_urls_from_string(post.selftext)
    # find references
    references = [parse_reddit_url(url) for url in urls]
    references = [r for r in references if r is not None]
    # ignore references to this specific post
    references = [r for r in references if (r["type"] != "post") or (r["post"] != post.id)]
    return references


def get_reddit_references_from_text(s):
    """
    Return all reddit references contained in a text.

    @param s: text to get references from
    @type s: L{str}
    @return: all references in the post, like L{parse_reddit_url}
    @rtype: L{list} of L{dict}
    """
    # get all URLs
    urls = get_urls_from_string(s)
    # find references
    references = [parse_reddit_url(url) for url in urls]
    references = [r for r in references if r is not None]
    return references


def has_post_locally(session, postid):
    """
    Check if a post exists locally.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param postid: id of post to check for
    @type postid: L{str}
    """
    stmt = select(Post.id).where(Post.id == postid)
    post = session.execute(stmt).one_or_none()
    return (post is not None)


def has_subreddit_locally(session, subreddit_name):
    """
    Check if a subreddit exists locally.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param subreddit_name: name of subreddit to check for
    @type subreddit_name: L{str}
    """
    stmt = select(Subreddit.name).where(Subreddit.name == subreddit_name)
    subreddit = session.execute(stmt).one_or_none()
    return (subreddit is not None)


def fetch_all_references(session, sleep=1):
    """
    Fetch all referenced objects, be it from crossposts or wiki references.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param sleep: how many seconds to wait between requests
    @type sleep: L{str}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_something_new = False
    # posts
    n = session.execute(select(func.count(Post.uid))).one()[0]
    stmt = select(Post).options(
        undefer(Post.url),
        undefer(Post.selftext),
    ).execution_options(
        yield_per=1000,
    )
    for post in tqdm.tqdm(session.execute(stmt).scalars(), desc="Searching in posts and fetching results...", total=n, unit="posts"):
        references = get_reddit_references_from_post(
            post,
        )
        references = [
            e
            for e in references
            if (e["type"] in ("post", "comment")) and not has_post_locally(session=session, postid=e["post"])
        ]
        if not references:
            continue
        for reference in tqdm.tqdm(references, desc="Fetching referenced objects for {}".format(post.id), total=len(references), unit="obj"):
            if reference["type"] in ("post", "comment"):
                if not has_post_locally(session, postid=reference["post"]):
                    # we need a second check in case the same post referenced the same post multiple times
                    fetch_post(session=session, postid=reference["post"], sleep=sleep)
                    did_fetch_something_new = True
            time.sleep(sleep)
    # wikipages
    n = session.execute(select(func.count(WikiPage.uid))).one()[0]
    stmt = select(WikiPage).options(
        undefer(WikiPage.content),
    ).execution_options(
        yield_per=1000,
    )
    for wikipage in tqdm.tqdm(session.execute(stmt).scalars(), desc="Searching in wikipages and fetching results...", total=n, unit="posts"):
        references = get_reddit_references_from_text(
            wikipage.content,
        )
        references = [
            e
            for e in references
            if (e["type"] in ("post", "comment")) and not has_post_locally(session=session, postid=e["post"])
        ]
        if not references:
            continue
        for reference in tqdm.tqdm(references, desc="Fetching referenced objects for {}/wiki/{}".format(wikipage.subreddit_name, wikipage.basepath), total=len(references), unit="obj"):
            if reference["type"] in ("post", "comment"):
                if not has_post_locally(session, postid=reference["post"]):
                    # we need a second check in case the same post referenced the same post multiple times
                    fetch_post(session=session, postid=reference["post"], sleep=sleep)
                    did_fetch_something_new = True
            time.sleep(sleep)
    return did_fetch_something_new


def fetch_post(session, postid, sleep=1):
    """
    Fetch a post from Arctic Shift and insert it into the databse.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param postid: id of post to check for
    @type postid: L{str}
    @param sleep: time to sleep between each request, in seconds
    @type sleep: L{int} or L{float}
    """
    # fetch the post
    url = "https://arctic-shift.photon-reddit.com/api/posts/ids?ids={}".format(postid)
    r = requests.get(url)
    r.raise_for_status()
    json = r.json()
    if ("data" not in json) or (len(json["data"]) == 0):
        return
    import_posts(session, json["data"])
    # fetch the comments
    start = int(json["data"][0].get("created_utc", 0))
    end = int(time.time())
    timestamp = start
    if timestamp > 0:
        timestamp - 100  # just to be sure we fetch all data
    raw_comments = []
    with tqdm.tqdm(
        desc="Fetching comments for {}".format(postid),
        total=(end - start),
        unit="seconds",
    ) as bar:
        while True:
            time.sleep(sleep)
            comment_url = "https://arctic-shift.photon-reddit.com/api/comments/search?link_id={}&sort=asc&after={}&limit=auto".format(
                postid,
                datetime.datetime.fromtimestamp(timestamp).isoformat(),
            )
            r = requests.get(comment_url)
            r.raise_for_status()
            json = r.json()
            if ("data" not in json) or (len(json["data"]) == 0):
                break
            raw_comments += json["data"]
            bar.set_postfix({"comments": len(raw_comments)})
            new_timestamp = timestamp
            for comment in json["data"]:
                if comment.get("created_utc", 0) > new_timestamp:
                    new_timestamp = int(comment["created_utc"])
            if new_timestamp == timestamp:
                new_timestamp += 1
            timestamp = new_timestamp
            bar.n = int(timestamp - start)
        bar.n = int(end - start)
        import_comments(session, raw_comments)


def get_wikipages_for_subreddit(subreddit_name):
    """
    Fetch wiki pages for a specific subreddit and return them.

    @param subreddit_name: name of subreddit to fetch wikipages for
    @type subreddit_name: L{str}
    @return: a list of wiki pages
    @rtype: L{list} of L{arcticzim.db.models.WikiPage}
    """
    url = "https://arctic-shift.photon-reddit.com/api/subreddits/wikis?subreddit={}&limit=100".format(subreddit_name)
    r = requests.get(url)
    r.raise_for_status()
    rawpages = r.json()["data"]
    pages = []
    for rawpage in rawpages:
        page = WikiPage(
            subreddit_name=subreddit_name,
            path=rawpage["path"],
            content=rawpage.get("content", "[page empty or content not available]"),
            revision_date=rawpage.get("revision_date", 0),
            revision_author=rawpage.get("revision_author", None) or "[Author unknown]",
            revision_reason=rawpage.get("revision_reason", None),
            retrieved_on=rawpage.get("retrieved_on", 0),
        )
        pages.append(page)
    return pages


def fetch_wiki_for_subreddit(session, subreddit_name):
    """
    Fetch wiki pages for a specific subreddit and insert them into the database.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param subreddit_name: name of subreddit to fetch wikipages for
    @type subreddit_name: L{str}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_something = False
    pages = get_wikipages_for_subreddit(subreddit_name)
    for page in pages:
        session.merge(page)
        did_fetch_something = True
    session.commit()
    return did_fetch_something


def fetch_all_wikis(session, sleep=1):
    """
    Fetch all wiki pages and insert them into the database.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param sleep: how many seconds to wait between each request
    @type sleep: L{int}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_something_new = False
    stmt = select(Subreddit).where(~Subreddit.wikipages.any())
    for subreddit in session.execute(stmt).scalars():
        print("Fetching wikipages for: {}".format(subreddit.name))
        if fetch_wiki_for_subreddit(session, subreddit.name):
            did_fetch_something_new = True
        time.sleep(sleep)
    return did_fetch_something_new


def get_rules_for_subreddit(subreddit_name):
    """
    Fetch the rules for a specific subreddit and return them.

    @param subreddit_name: name of subreddit to fetch rules for
    @type subreddit_name: L{str}
    @return: a list of subreddit rules
    @rtype: L{list} of L{arcticzim.db.models.SubredditRule}
    """
    url = "https://arctic-shift.photon-reddit.com/api/subreddits/rules?subreddits={}".format(subreddit_name)
    r = requests.get(url)
    r.raise_for_status()
    all_rawrules = r.json()["data"]
    if not all_rawrules:
        # no rules for subreddits (found)
        return []
    rawrules = all_rawrules[0]["rules"]
    rules = []
    for rawrule in rawrules:
        rule = SubredditRule(
            subreddit_name=subreddit_name,
            kind=rawrule["kind"],
            priority=rawrule["priority"],
            short_name=rawrule["short_name"],
            created_utc=rawrule["created_utc"],
            description=rawrule["description"],
            violation_reason=rawrule.get("violation_reason", ""),
        )
        rules.append(rule)
    return rules


def fetch_rules_for_subreddit(session, subreddit_name):
    """
    Fetch the rules for a specific subreddit and insert them into the database.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param subreddit_name: name of subreddit to fetch rules for
    @type subreddit_name: L{str}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_something = False
    rules = get_rules_for_subreddit(subreddit_name)
    for rule in rules:
        session.merge(rule)
        did_fetch_something = True
    session.commit()
    return did_fetch_something


def fetch_all_rules(session, sleep=1):
    """
    Fetch all rules and insert them into the database.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param sleep: how many seconds to wait between each request
    @type sleep: L{int}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_something_new = False
    stmt = select(Subreddit).where(~Subreddit.rules.any())
    for subreddit in session.execute(stmt).scalars():
        print("Fetching rules for: {}".format(subreddit.name))
        if fetch_rules_for_subreddit(session, subreddit.name):
            did_fetch_something_new = True
        time.sleep(sleep)
    return did_fetch_something_new


def fetch_all(session, sleep=1, with_references=True):
    """
    Run all fetch operations.

    @param session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    @param sleep: how many seconds to wait between each request
    @type sleep: L{int}
    @param with_references: wether referenced reddit pages (e.g. posts) should also be fetched
    @type with_references: L{bool}
    @return: whether anything new has been fetched
    @rtype: L{bool}
    """
    did_fetch_wiki = fetch_all_wikis(session, sleep=sleep)
    did_fetch_rule = fetch_all_rules(session, sleep=sleep)
    if with_references:
        did_fetch_post = fetch_all_references(session, sleep=sleep)
    return any((did_fetch_wiki, did_fetch_rule, did_fetch_post))


class ReferenceUrlRewriter(object):
    """
    This class handles the rewriting of URLs referencing reddit pages.

    @ivar session: sqlalchemy session to use
    @type session: L{sqlalchemy.orm.Session}
    """
    def __init__(self, session):
        """
        The default constructor.

        @param session: sqlalchemy session to use
        @type session: L{sqlalchemy.orm.Session}
        """
        self.session = session

    def should_rewrite(self, reference):
        """
        Check if a reference should be rewritten.

        @param reference: reference to check
        @type reference: L{dict}
        @return: whether the reference should be rewritten or not
        @rtype: L{bool}
        """
        if reference["type"] in ("post", "comment"):
            return has_post_locally(self.session, reference["post"])
        elif reference["type"] == "subreddit":
            return has_subreddit_locally(self.session, reference["subreddit"])
        else:
            return False

    def rewrite_url(self, url, to_root):
        """
        Rewrites an external URL to an internal one if the referenced object exists locally.

        @param url: url to rewrite
        @type url: L{str}
        @param to_root: like with the renderer, a prefix (e.g. C{../..}) that indicates how to navigate to the root directory
        @type to_root: L{str}
        @return: the url to use, which may still be an external one
        @rtype: L{str}
        """
        reference = parse_reddit_url(url)
        if reference is None:
            # not a reddit url
            return url
        if self.should_rewrite(reference):
            return reddit_reference_to_url(reference, to_root=to_root)
        else:
            return url

    def rewrite_urls_in_text(self, text, to_root):
        """
        Rewrite all external URLs found in the text with internal ones if the files exist locally.

        @param text: text to rewrite
        @type text: L{str}
        @param to_root: like with the renderer, a prefix (e.g. C{../..}) that indicates how to navigate to the root directory
        @type to_root: L{str}
        @return: the rewritten text
        @rtype: L{str}
        """
        urls = get_urls_from_string(text)
        for url in urls:
            new_url = self.rewrite_url(url, to_root=to_root)
            if new_url != url:
                text = text.replace(url, new_url)
        return text



"""
Image-related utilities.
"""
import argparse
import math
import os

from PIL import Image


def mimetype_is_image(mimetype):
    """
    Check if a mimetype refers to an image mimetype.

    @param mimetype: mimetype to check
    @type mimetype: L{bool}
    @return: whether the mimetype is an image mimetype
    @rtype: L{bool}
    """
    if ";" in mimetype:
        mimetype = mimetype[:mimetype.find(";")]
    mimetype = mimetype.strip().lower()
    return mimetype.startswith("image/")


def mimetype_is_video(mimetype):
    """
    Check if a mimetype refers to a video mimetype.

    @param mimetype: mimetype to check
    @type mimetype: L{bool}
    @return: whether the mimetype is a video mimetype
    @rtype: L{bool}
    """
    if ";" in mimetype:
        mimetype = mimetype[:mimetype.find(";")]
    mimetype = mimetype.strip().lower()
    return mimetype.startswith("video/")


def minimize_image(path, max_w=512, max_h=512):
    """
    Minimize the image at the target path.

    @param path: path to image which should be minimized
    @type path: L{str}
    @param max_w: max width for the new image
    @type max_w: L{int}
    @param max_h: max height for the new image
    @type max_h: L{int}
    @return: a tuple of (new_mimetype, new_size)
    @rtype: L{tuple} of (L{str}, L{int})
    """
    with Image.open(path) as img:
        # load image and close the file
        img.load()
        # resize
        w_ratio = (max_w / img.width)
        h_ratio = (max_h / img.height)
        ratio = min(w_ratio, h_ratio, 1)
        new_w = math.floor(img.width * ratio)
        new_h = math.floor(img.height * ratio)
        new_img = img.resize((new_w, new_h))
    with open(path, "wb") as fout:
        # convert and save
        new_img.save(fout, "WEBP")
        fout.seek(0, os.SEEK_END)
        size = fout.tell()
    return ("image/webp", size)


if __name__ == "__main__":
    """
    Test code.
    """
    parser = argparse.ArgumentParser(description="Minimize an image in-place")
    parser.add_argument("path", action="store", help="path to image")
    ns = parser.parse_args()
    minimize_image(ns.path)



"""
Utilities for working with jsonl files.
"""
import argparse
import os
import json
import pprint
import math
import datetime

from tqdm.auto import tqdm


def process_jsonl(path, desc="Reading file"):
    """
    Process a jsonl file, yielding each element.

    @param path: path to the file to read
    @type path: L{str}
    @param desc: description for the tqdm progressbar
    @type desc: L{str}
    @yields: each element in the jsonl file
    @ytype: a json element, usually a L{dict}
    """
    with open(path, "r") as fin:
        total_size = fin.seek(0, os.SEEK_END)
        fin.seek(0, os.SEEK_SET)
        entry = 0

        with tqdm(desc=desc, total=total_size, unit="B", unit_scale=True, unit_divisor=1024) as t:

            for line in fin:
                entry += 1
                t.set_postfix(entry=entry, refresh=False)
                t.update(len(line))
                sline = line.strip()
                if sline:
                    yield json.loads(sline)


def analyze_jsonl(path):
    """
    Analyze the contents of a jsonl file, returning info about the keys and values.

    @param path: path to the file to analyze
    @type path: L{str}
    @return: details about the data
    @rtype: L{dict}
    """
    fields = {}
    first = True
    for element in process_jsonl(path=path, desc="Analyzing file"):
        keys = element.keys()
        for key in keys:
            value = element[key]
            if key not in fields:
                fields[key] = {
                    "always_present": first,
                    "types": set(),
                    "nullable": False,
                    "example": value,
                    "count": 0,
                }
            fields[key]["types"].add(type(value))
            fields[key]["count"] += 1
            if value and not fields[key]["example"]:
                fields[key]["example"] = value
            if value is None:
                fields[key]["nullable"] = True
            if isinstance(value, (str, list, tuple)):
                length = len(value)
                if ("max_length" not in fields[key]) or (fields[key]["max_length"] < length):
                    fields[key]["max_length"] = length
            if isinstance(value, (int, float)):
                if ("min_value" not in fields[key]) or (fields[key]["min_value"] > value):
                    fields[key]["min_value"] = value
                if ("max_value" not in fields[key]) or (fields[key]["max_value"] < value):
                    fields[key]["max_value"] = value

            for existing_key in fields:
                if existing_key not in keys:
                    fields[existing_key]["always_present"] = False
        first = False

    return fields


def generate_columns(fields):
    """
    Generate columns for a jsonl.

    @param fields: field analysis as returned by L{analyze_jsonl}
    @type fields: L{str}
    @return: a list of field definitions
    @rtype: L{list} of L{str}
    """
    defs = []
    for fieldname in sorted(list(fields.keys())):
        field = fields[fieldname]
        # get field type
        fieldtypes = field["types"]
        fieldtypedef = None
        if str in fieldtypes:
            fieldtypedef = "Unicode({})".format(
                2 ** math.ceil(math.log(field["max_length"], 2)) if field["max_length"] > 0 else 128,
            )
            used_fieldtype = str
        elif bool in fieldtypes:
            # fieldtypedef = "Boolean"
            used_fieldtype = bool
        elif int in fieldtypes:
            # fieldtypedef = "Integer"
            used_fieldtype = int
        elif datetime.datetime in fieldtypes:
            # fieldtypedef = "DateTime"
            used_fieldtype = datetime.datetime
        else:
            used_fieldtype = None
        # nullable check
        nullable = field["nullable"] or not field["always_present"]
        if used_fieldtype is not None:
            fielddef = "{}: Mapped[{}]{}".format(
                fieldname,
                (used_fieldtype.__name__ if not nullable else "Optional[{}]".format(used_fieldtype.__name__)),
                (" = mapped_column({})".format(fieldtypedef) if fieldtypedef is not None else ""),
            )
        else:
            fielddef = "# {}: {}".format(fieldname, fields[fieldname])
        defs.append(fielddef)
    return defs


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Examine a jsonl file.")
    parser.add_argument("path", help="path to file to read")
    parser.add_argument(
        "--generate-column-definitions",
        action="store_true",
        dest="generate_columns",
        help="If set, output sqlalchemy column definitions instead",
    )
    ns = parser.parse_args()

    result = analyze_jsonl(ns.path)
    if ns.generate_columns:
        for fielddef in generate_columns(result):
            print(fielddef)
    else:
        pprint.pprint(result)



"""
Various utility functions.

@var ALLOWED_WORD_LETTERS: a regular expression pattern used to to identify where words end.
@type ALLOWED_WORD_LETERS: L{re.Pattern}
@var URL: a regular expression matching likely URLs
@type URL: L{re.Pattern}
"""
import datetime
import re
import os
import decimal
from urllib.parse import urlparse


ALLOWED_WORD_LETTERS = re.compile(r"[^\w|\-]")
# from https://stackoverflow.com/a/3809435 (modified)
URL = re.compile(r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*[-a-zA-Z0-9@:%_\+.~#?&//=])")


def format_timedelta(seconds):
    """
    Format seconds since event into a readable string.

    @param seconds: seconds to format
    @type seconds: L{int}
    @return: the formated string
    @rtype: L{str}
    """

    formatted = str(datetime.timedelta(seconds=seconds))
    if "." in formatted:
        formatted = formatted[:-4]
    return formatted


def format_number(n, allow_none=False):
    """
    Format a number to be a bit more human readable.

    @param n: number to format
    @type n: L{int}
    @param allow_none: if false, raise an exception on None values
    @type allow_none: L{bool}
    @return: the formated string
    @rtype: L{str}
    """
    if isinstance(n, decimal.Decimal):
        n = float(n)
    if n is None:
        if allow_none:
            return "-"
        else:
            raise TypeError("format_number() got called with 'None' and allow_none=False!")
    if n < 1000 and isinstance(n, int):
        return str(n)
    for fmt in ("", "K", "M", "B", "T", "Qa"):
        if n < 1000.0:
            return "{:.2f}{}".format(round(n, 3), fmt)
        else:
            n /= 1000.0
    return "{:.2f}Qi".format(round(n, 2))


def format_size(nbytes):
    """
    Format the given byte count into a human readable format.

    @param nbytes: size in bytes
    @type nbytes: L{int}
    @return: a human readable string describing the size
    @rtype: L{str}
    """
    for fmt in ("B", "KiB", "MiB", "GiB", "TiB", "PiB"):
        if nbytes < 1024.0:
            return "{:.2f} {}".format(round(nbytes, 2), fmt)
        else:
            nbytes /= 1024.0
    return "{:.2f} EiB".format(round(nbytes, 2))


def format_date(date):
    """
    Format a date.

    @param date: date to format
    @type date: L{datetime.datetime}
    @return: the formated date
    @rtype: L{str}
    """
    return date.strftime("%Y-%m-%d")


def get_package_dir():
    """
    Return the path to the root directory of this package.

    @return: the path to the root directory of this package (not repo!)
    @rtype: L{str}
    """
    return os.path.dirname(__file__)


def get_resource_file_path(*names):
    """
    Return the path to the specified resource file.

    @param names: name(s) of the resource file to get path for. If multiple are specified, they are interpeted as a sequence of path segments.
    @type names: L{str}
    @return: path to the resource file
    @rtype: L{str}
    """
    p = os.path.join(get_package_dir(), "resources", *names)
    return p


def add_to_dict_list(d, k, v):
    """
    Add a value to a dictionary of lists.

    If k is in d, append v. Otherwise, set d[k] = [v].

    @param d: dictionary of lists to append v to
    @type d: L{dict} of hashable -> L{list}
    @param k: key of list to append v to
    @type k: hashable
    @param v: value to append
    @type v: any
    """
    if k in d:
        d[k].append(v)
    else:
        d[k] = [v]


def count_words(text):
    """
    Count the words in the text.

    @param text: text to count words in.
    @type text: L{str}
    @return: number of words in text.
    @rtype: L{int}
    """
    return len(ALLOWED_WORD_LETTERS.sub(" ", text).split())


def set_or_increment(d, k, v=1):
    """
    Set or increment a key in a dict to/by a value.

    Basically, if k in d set d[k] += v, else d[k] = v.

    @param d: dictionary to modify
    @type d: L{dict}
    @param k: key in dict to use
    @type k: hashable
    @type v: value to set to or increment by
    @type v: L{int} or L{float}
    """
    if k in d:
        d[k] += v
    else:
        d[k] = v


def delete_or_decrement(d, k, v=1, delete_on=1):
    """
    Delete or increment a key in a dict for/by a value.

    Basically, if d[k] >= delete_on set d[k] -= v else del d[k]

    @param d: dictionary to modify
    @type d: L{dict}
    @param k: key in dict to use
    @type k: hashable
    @type v: value to decrement by
    @type v: L{int} or L{float}
    @param delete_on: if prior to decrement the value is <= this value, delete it
    @raises KeyError: if k not in d
    """
    if d[k] <= delete_on:
        del d[k]
    else:
        d[k] -= v


def ensure_iterable(obj):
    """
    If obj is iterable, return obj, else return an iterable yielding obj.

    May not work correctly on primitive data types.

    @param obj: object to turn iterable
    @type obj: any
    @return: an iterable (either obj or one yielding obj)
    @rtype: iterable
    """
    if hasattr(obj, "__iter__"):
        return obj
    else:
        return (obj, )


def remove_duplicates(li):
    """
    Copy a list such that the order of elements is preserved but only
    the first occurrence of each element preserved.

    @param li: list to sort
    @type li: L{list}
    @return: a copy of the list with some elements potentially removed
    @rtype: L{list}
    """
    ret = []
    for e in li:
        if e not in ret:
            ret.append(e)
    return ret


def chunked(iterable, n):
    """
    Split an iterable into multiple lists, each containing at most n elements.

    @param iterable: iterable to split
    @type iterable: iterable
    @param n: number of elements each list should have at most
    @type n: L{int}
    @return: a generator yielding lists, each a chunk of the input data
    @rtype: generator yielding L{list}
    """
    current = []
    for element in iterable:
        current.append(element)
        if len(current) >= n:
            yield current
            current = []
    if current:
        yield current


def get_urls_from_string(s):
    """
    Find all URLs in a string.

    No guarantees are made about the validity or completenes of the result.

    @param s: string to search for URLs
    @type s: L{str}
    @return: a list of URLs
    @rtype: L{str}
    """
    return [s[m.start():m.end()] for m in URL.finditer(s)]


def parse_reddit_url(url):
    """
    Parse a reddit URL.

    The result is a dict containing some info, most importantly the "type" key.

    @param url: URL to parse
    @type url: L{str}
    @return: info about the URL if it is a reddit url, otherwise L{None}
    @rtype: L{dict} or L{None}
    """
    if isinstance(url, bytes):
        url = url.decode("utf-8")
    if url == "":
        return None
    parts = urlparse(url)
    host = parts.hostname
    if (host is None) or not (host.endswith("reddit.com") or host.endswith("redd.it")):
        # not a reddit link
        return None
    path = parts.path
    segments = path.split("/")
    if segments and not segments[0]:
        segments.pop(0)
    if segments and not segments[-1]:
        segments.pop(-1)
    if len(segments) == 0:
        # just a general reddit url
        return None
    if segments[0] in ("u", "user"):
        # a user reference
        username = segments[1]
        return {"scheme": parts.scheme, "type": "user", "username": username}
    elif segments[0] == "r":
        # a subreddit, post or comment reference
        subredditname = segments[1]
        if (len(segments) >= 6) and (segments[2] == "comments"):
            # a comment reference
            return {"scheme": parts.scheme, "type": "comment", "subreddit": subredditname, "post": segments[3], "comment": segments[5]}
        if (len(segments) >= 4) and (segments[2] == "comments"):
            # a post reference
            return {"scheme": parts.scheme, "type": "post", "subreddit": subredditname, "post": segments[3]}
        else:
            # just a subreddit reference
            return {"scheme": parts.scheme, "type": "subreddit", "subreddit": subredditname}
    else:
        # unknown reddit reference
        return None


def trim_title(s):
    """
    Trim a title, making it more suitable for inclusion in a ZIM file.

    @param s: title to trim
    @type s: L{str}
    @return: the trimemd title
    @rtype: L{str}
    """
    words = s.split(" ")
    words = [trim_word(word) for word in words]
    title = " ".join(words)
    if len(title) > 230:
      title = title[:230]
    return title


def trim_word(word):
    """
    Trim a word, making it more suitable in titles.

    @param word: word to trim
    @type word: L{str}
    @return: the trimmed word
    @rtype: L{str}
    """
    max_word_length = 100
    if len(word) >= max_word_length:
        word = word[:max_word_length]
    word = word.strip()
    return word


if __name__ == "__main__":
    # test code
    val = int(input("n: "))
    print("Formated as number:   {}".format(format_number(val)))
    print("Formated as size:     {}".format(format_size(val)))
    print("Formated as timedela: {}".format(format_timedelta(val)))



"""
This module contains the command line interface.
"""
import argparse
import time
import os


try:
    import multiprocessing
    multiprocessing.set_start_method("forkserver")
except Exception:
    # multiprocessing may not be available
    multiprocessing = None

from sqlalchemy.orm import Session

from .db.models import Base
from .db.connection import ConnectionConfig
from .importer import import_posts_from_file, import_comments_from_file, prepare_db
from .zimbuild.builder import ZimBuilder, BuildOptions
from .downloader import download_all as download_all_media
from .fetcher import fetch_all
from .util import format_timedelta


def _connection_config_from_ns(ns):
    """
    Generate a connection configuration from the argparse namespace.

    @param ns: namespace containing arguments
    @type ns: L{argparse.Namespace}
    @return: the connection config
    @rtype: L{arcticzim.db.connection.ConnectionConfig}
    """
    config = ConnectionConfig(
        url=ns.database,
        verbose=(ns.verbose >= 2),
    )
    return config


def run_import(ns):
    """
    Run the import command.

    @param ns: namespace containing arguments
    @type ns: L{argparse.Namespace}
    """
    start = time.time()
    print("Connecting to database...")
    engine = _connection_config_from_ns(ns).connect()
    print("Database connection established. Creating tables if necessary...")
    Base.metadata.create_all(engine, checkfirst=True)
    print("Done. Creating databse session...")
    with Session(engine) as session:
        print("Done. Preparing database...")
        prepare_db(session)
        print("Done. Importing posts...")
        import_posts_from_file(session, path=ns.posts_file, batch_size=ns.batch_size)
        print("Done. Importing comments..")
        import_comments_from_file(session, path=ns.comments_file, batch_size=ns.batch_size)
    print("Import finished in  {}".format(format_timedelta(time.time() - start)))


def run_fetch(ns):
    """
    Run the fetch-extra command.

    @param ns: namespace containing arguments
    @type ns: L{argparse.Namespace}
    """
    start = time.time()
    print("Connecting to database...")
    engine = _connection_config_from_ns(ns).connect()
    print("Database connection established. Creating tables if necessary...")
    Base.metadata.create_all(engine, checkfirst=True)
    print("Done. Creating databse session...")
    with Session(engine) as session:
        print("Done. Starting fetch...")
        rnd = 1
        while True:
            print("Fetch round #{}".format(rnd))
            did_fetch_something = fetch_all(session, sleep=ns.sleep)
            if ns.single:
                print("--single specified, stopping after first fetch round")
                break
            if did_fetch_something:
                print("Fetched something new, starting another fetch round in case some new references have been added.")
                rnd += 1
            else:
                print("Did not fetch anything new.")
                break
    print("Fetch finished in  {}".format(format_timedelta(time.time() - start)))


def run_build(ns):
    """
    Run the build command.

    @param ns: namespace containing arguments
    @type ns: L{argparse.Namespace}
    """
    connection_config = _connection_config_from_ns(ns)
    builder = ZimBuilder(connection_config, mediadir=ns.mediadir)
    build_options = BuildOptions.from_ns(ns)
    builder.build(ns.outpath, options=build_options)


def run_media_download(ns):
    """
    Run the download-media command.

    @param ns: namespace containing arguments
    @type ns: L{argparse.Namespace}
    """
    print("Creating media directory if neccessary...")
    if not os.path.exists(ns.mediadir):
        os.mkdir(ns.mediadir)
    connection_config = _connection_config_from_ns(ns)
    print("Done. Connecting to database...")
    engine = connection_config.connect()
    print("Done. Creating database session...")
    with Session(engine) as session:
        print("Done. Starting download.")
        download_all_media(
            session=session,
            mediadir=ns.mediadir,
            enable_post_processing=ns.post_processing,
            download_reddit_videos=ns.download_reddit_videos,
            download_external_videos=ns.download_external_videos,
            max_image_dimension=ns.max_image_dimension,
            include_comments=ns.include_comments,
        )


def main():
    """
    The main function.
    """
    parser = argparse.ArgumentParser(description="Create ZIM files from arcticshift.")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="be more verbose",
    )
    subparsers = parser.add_subparsers(
        dest="command",
        required=True,
        help="command to execute",
    )

    # parser for the import
    import_parser = subparsers.add_parser(
        "import",
        help="Import a subreddit into a database.",
    )
    import_parser.add_argument(
        "database",
        action="store",
        help="Database to store objects in, as sqlalchemy connection URL",
    )
    import_parser.add_argument(
        "--posts-file",
        action="store",
        dest="posts_file",
        help="Path to file containing post data to import from",
    )
    import_parser.add_argument(
        "--comments-file",
        action="store",
        dest="comments_file",
        help="Path to file containing comment data to import from",
    )
    import_parser.add_argument(
        "--batch-size",
        action="store",
        type=int,
        dest="batch_size",
        default=1000,
        help="how many posts and comments to import at once",
    )

    fetch_parser = subparsers.add_parser(
        "fetch-extra",
        help="fetch additional data (e.g. wiki pages)",
    )
    fetch_parser.add_argument(
        "database",
        action="store",
        help="database to load objects from, as sqlalchemy connection URL",
    )
    fetch_parser.add_argument(
        "--sleep",
        action="store",
        type=float,
        default=1,
        help="how many seconds to wait between requests",
    )
    fetch_parser.add_argument(
        "--single",
        action="store_true",
        help="Perform at most a single fetch round.",
    )

    mediadownload_parser = subparsers.add_parser(
        "download-media",
        help="download media files of posts",
    )
    mediadownload_parser.add_argument(
        "database",
        action="store",
        help="database to load objects from, as sqlalchemy connection URL",
    )
    mediadownload_parser.add_argument(
        "--media-dir",
        action="store",
        dest="mediadir",
        default="arcticzim_media/",
        help="directory to store media in",
    )
    mediadownload_parser.add_argument(
        "--no-post-processing",
        action="store_false",
        dest="post_processing",
        help="disable post-processing, keeping media files unaltered.",
    )
    mediadownload_parser.add_argument(
        "--download-reddit-videos",
        action="store_true",
        dest="download_reddit_videos",
        help="download videos hosted on reddit",
    )
    mediadownload_parser.add_argument(
        "--download-external-videos",
        action="store_true",
        dest="download_external_videos",
        help="download videos NOT hosted on reddit",
    )
    mediadownload_parser.add_argument(
        "--max-image-dimension",
        action="store",
        type=int,
        dest="max_image_dimension",
        default=512,
        help="Downscale images to this many pixels on their longer side",
    )
    mediadownload_parser.add_argument(
        "--search-comments",
        action="store_true",
        dest="include_comments",
        help="also download media linked in comments",
    )

    # parser for the ZIM build
    build_parser = subparsers.add_parser(
        "build",
        help="build a ZIM file",
    )
    build_parser.add_argument(
        "database",
        action="store",
        help="database to load objects from, as sqlalchemy connection URL",
    )
    build_parser.add_argument(
        "--media-dir",
        action="store",
        dest="mediadir",
        default="arcticzim_media/",
        help="directory to store media in",
    )
    build_parser.add_argument(
        "outpath",
        action="store",
        help="path to write ZIM to",
    )
    BuildOptions.add_argparse_options(build_parser)

    ns = parser.parse_args()

    if ns.command == "import":
        run_import(ns)
    elif ns.command == "fetch-extra":
        run_fetch(ns)
    elif ns.command == "download-media":
        run_media_download(ns)
    elif ns.command == "build":
        run_build(ns)
    else:
        raise RuntimeError("Unknown subcommand: {}".format(ns.command))


if __name__ == "__main__":
    main()



"""
This module contains the mistune (markdown) customization.
"""
import mistune
from mistune.util import escape_url


class CustomMistuneBlockLevelParser(mistune.BlockParser):
    """
    A custom block-level parser for mistune that adjusts some behavior for reddit.

    More specifically, this one allows us to not have a space before a title.
    """
    SPECIFICATION = mistune.BlockParser.SPECIFICATION.copy()


CustomMistuneBlockLevelParser.SPECIFICATION["atx_heading"] = r"^ {0,3}(?P<atx_1>#{1,6})(?!#+)(?P<atx_2>[ \t]*|[ \t]*.*?)$"


URL_LINK_PATTERN = r"(\.\./)+[^\s<]+[^<.,:;\"')\]\s]"


def parse_url_link(inline, m, state):
    text = m.group(0)
    pos = m.end()
    if state.in_link:
        inline.process_text(text, state)
        return pos
    state.append_token(
        {
            "type": "link",
            "children": [{"type": "text", "raw": text}],
            "attrs": {"url": escape_url(text)},
        }
    )
    return pos


def relative_url_plugin(md):
    md.inline.register("url_link", URL_LINK_PATTERN, parse_url_link)



"""
Helper code for debugging workers.
"""
import queue
import cmd
import pdb
import shlex
import argparse
import time

from sqlalchemy import create_engine

from .worker import Worker, WorkerOptions
from .worker import PostRenderTask, EtcRenderTask, SubredditRenderTask, UserRenderTask
from .renderer import RenderOptions
from ..util import ensure_iterable


class DiscardingQueue(queue.Queue):
    """
    A type of L{queue.Queue} which discards elements being put into it.

    @ivar verbose: if nonzero, print items put into queue
    @type verbose: L{bool}
    """
    def __init__(self, *args, verbose=False, **kwargs):
        """
        The default constructor.

        @param verbose: if nonzero, print items put into queue
        @type verbose: L{bool}
        """
        queue.Queue.__init__(self, *args, **kwargs)
        self.verbose = verbose

    def qsize(self):
        return 0

    def empty(self):
        return True

    def full(self):
        return False

    def put(self, item, block=True, timeout=None):
        if self.verbose:
            print(str(item))

    def put_nowait(self, item):
        if self.verbose:
            print(str(item))


class DebugWorker(Worker):
    """
    A variant of L{arcticzim.zimbuild.worker.Worker} for debugging purposes.

    This worker is intended to be used standalone, without the builder.
    """
    def __init__(self, parent, engine, options, render_options):
        """
        Like L{arcticzim.zimbuild.worker.Worker.__init__}, but with some
        fixed options. Additionally, the 'parent' argument must be provided,
        pointing to the associated L{WorkerDebugger}.

        @param parent: the associated worker debugger
        @type parent: L{WorkerDebugger}
        """
        Worker.__init__(
            self,
            id=0,
            inqueue=queue.Queue(),
            outqueue=DiscardingQueue(verbose=parent.verbose),
            engine=engine,
            options=options,
            render_options=render_options,
        )
        self.parent = parent

    def setup_logging(self):
        # do not setup an output file
        self._last_log_time = time.time()
        self._log_file = None

    def log(self, msg):
        assert isinstance(msg, str)
        full_msg = "[{}][+{:8.3f}s] {}\n".format(
            time.ctime(),
            time.time() - self._last_log_time,
            msg,
        )
        print(full_msg, end="")
        self._last_log_time = time.time()

    def handle_result(self, result):
        it = ensure_iterable(result)
        for i, subresult in enumerate(it):
            self.log("Submitting result part {}...".format(i))
            self.outqueue.put(subresult)
            # inform the parent that we've got a result part
            self.parent.got_result_part(i, subresult)


class WorkerDebugger(cmd.Cmd):
    """
    A CLI for debugging the worker.

    @ivar worker: the worker used
    @type worker: L{DebugWorker}
    @ivar engine: engine used for database connection
    @type engine: L{sqlalchemy.engine.Engine}
    @ivar verbose: if nonzero, be more verbose
    @type verbose: L{bool}
    """
    def __init__(self, engine, options, verbose=False):
        """
        The default constructor.

        @param engine: engine used for database connection
        @type engine: L{sqlalchemy.engine.Engine}
        @param options: options for the worker
        @type options: L{arcticzim.zimbuild.worker.WorkerOptions}
        @param verbose: if nonzero, be more verbose
        @type verbose: L{bool}
        """
        cmd.Cmd.__init__(self)

        self.engine = engine
        self.verbose = verbose
        self.worker = DebugWorker(
            parent=self,
            engine=engine,
            options=options,
            render_options=RenderOptions(),
        )

        # states for debugging targets
        self.launch_pdb_after_result_part = None

    def got_result_part(self, i, part):
        """
        Called when the worker finished submitting a result.

        @param i: index of the part
        @type i: L{int}
        @param part: render result that was received
        @type part: L{arcticzim.zimbuild.renderer.RenderResult}
        """
        if self.launch_pdb_after_result_part is not None:
            if i == self.launch_pdb_after_result_part:
                pdb.set_trace()

    def do_quit(self, line):
        """
        quit: exit this CLI.
        """
        return True

    do_q = do_close = do_exit = do_quit

    def do_launch_pdb_after_subresult(self, line):
        """
        launch_pdb_after_subresult <i|disable>: launch pdb after a subresult with this index has been received
        """
        if line == "disable":
            self.launch_pdb_after_result_part = None
        else:
            try:
                i = int(line)
            except ValueError:
                print("Not a valid index: ", line)
                return
            self.launch_pdb_after_result_part = i

    def do_process_task(self, line):
        """
        process_task <task spec...>: create a task and tell the worker to process it.
        """
        splitted = shlex.split(line)
        if len(splitted) == 0:
            print("Error: empty task spec!")
            return
        tasktype = splitted.pop(0)
        if tasktype in ():
            task_uid = int(splitted[0])
        elif tasktype in ("post", ):
            task_uid = [int(e) for e in splitted[0].split(",")]
        elif tasktype in ("etc", "subreddit"):
            task_uid = splitted[0]
        elif tasktype in ("user"):
            task_uid = splitted[0].split("-")
            task_uid = (int(task_uid[0]), task_uid[1], task_uid[2])
        task_classes_and_functions = {
            "post": (PostRenderTask, self.worker.process_post_task),
            "subreddit": (SubredditRenderTask, self.worker.process_subreddit_task),
            "user": (UserRenderTask, self.worker.process_user_task),
            "etc": (EtcRenderTask, self.worker.process_etc_task),
        }
        cls, f = task_classes_and_functions[tasktype]
        task = cls(*task_uid)
        f(task)


def main():
    """
    The main function.
    """
    parser = argparse.ArgumentParser(description="Debug helper for the workers")
    parser.add_argument("db", help="sqlalchemy database URL")
    parser.add_argument("-v", "--verbose", action="store_true", help="be more verbpse")
    parser.add_argument("--lazy", action="store_false", dest="eager", help="load objects lazily")
    ns = parser.parse_args()

    engine = create_engine(ns.db, echo=ns.verbose)
    options = WorkerOptions(eager=ns.eager)
    debugger = WorkerDebugger(engine, options=options, verbose=ns.verbose)
    debugger.cmdloop()


if __name__ == "__main__":
    main()



"""
Database model definitions.

@var ARCTICZIM_USERNAME: a username used as author for some helper objects
@type ARCTICZIM_USERNAME: L{str}
"""
import datetime

from typing import List, Optional

from sqlalchemy import ForeignKey, UniqueConstraint
from sqlalchemy import String, Unicode
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from ..util import parse_reddit_url


ARCTICZIM_USERNAME = "_ArcticZim"


class Base(DeclarativeBase):
    """
    Sqlalchemy declarative base.
    """
    pass


class Subreddit(Base):
    """
    This model represents a subreddit in the database.
    """
    __tablename__ = "subreddit"
    name: Mapped[str] = mapped_column(Unicode(32), primary_key=True)
    subscribers: Mapped[int]

    posts: Mapped[List["Post"]] = relationship(
        back_populates="subreddit",
        cascade="all, delete-orphan",
    )
    comments: Mapped[List["Comment"]] = relationship(
        back_populates="subreddit",
        cascade="all, delete-orphan",
    )
    wikipages: Mapped[List["WikiPage"]] = relationship(
        back_populates="subreddit",
        cascade="all, delete-orphan",
    )
    rules: Mapped[List["SubredditRule"]] = relationship(
        back_populates="subreddit",
        cascade="all, delete-orphan",
    )


class User(Base):
    """
    This model represents a user (aka author) in the database.
    """
    __tablename__ = "user"
    __table_args__ = (
        UniqueConstraint("name"),
    )

    # uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    # id: Mapped[str] = mapped_column(String(16))
    # name: Mapped[str] = mapped_column(String(32), index=True, unique=True)
    name: Mapped[str] = mapped_column(String(32), primary_key=True)
    created: Mapped[datetime.datetime]

    posts: Mapped[List["Post"]] = relationship(
        back_populates="author",
        cascade="all, delete-orphan",
        # primaryjoin="User.name == Post.author",
    )
    comments: Mapped[List["Comment"]] = relationship(
        back_populates="author",
        cascade="all, delete-orphan",
        # primaryjoin="User.name == Post.author",
    )


class Post(Base):
    """
    This model represents a post in the database.
    """
    __tablename__ = "post"

    __table_args__ = (
        UniqueConstraint("id"),
        UniqueConstraint("name"),
    )

    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    id: Mapped[str] = mapped_column(String(8), index=True, unique=True)

    archived: Mapped[Optional[bool]]
    author_name: Mapped[str] = mapped_column(ForeignKey("user.name"), index=True)
    author_cakeday: Mapped[Optional[bool]]
    author_flair_background_color: Mapped[Optional[str]] = mapped_column(Unicode(8))
    author_flair_css_class: Mapped[Optional[str]] = mapped_column(Unicode(16))
    # author_flair_richtext: {'always_present': False, 'types': {<class 'list'>}, 'nullable': False, 'example': [], 'count': 55284, 'max_length': 0}
    author_flair_text: Mapped[Optional[str]] = mapped_column(Unicode(128))
    author_flair_text_color: Mapped[Optional[str]] = mapped_column(Unicode(16))
    author_flair_type: Mapped[Optional[str]] = mapped_column(Unicode(8))
    author_fullname: Mapped[Optional[str]] = mapped_column(Unicode(16))
    created_utc: Mapped[int]
    crosspost_parent: Mapped[Optional[str]] = mapped_column(String(16))
    distinguished: Mapped[Optional[str]] = mapped_column(String(16))
    domain: Mapped[Optional[str]] = mapped_column(Unicode(128))
    edited: Mapped[Optional[int]]
    # edited: Mapped[Optional[bool]]
    hide_score: Mapped[Optional[bool]]
    is_gallery: Mapped[Optional[bool]]
    is_meta: Mapped[Optional[bool]]
    is_original_content: Mapped[Optional[bool]]
    is_reddit_media_domain: Mapped[Optional[bool]]
    is_self: Mapped[bool]
    is_video: Mapped[Optional[bool]]
    link_flair_background_color: Mapped[Optional[str]] = mapped_column(Unicode(8))
    link_flair_css_class: Mapped[Optional[str]] = mapped_column(Unicode(16))
    # link_flair_richtext: {'always_present': False, 'types': {<class 'list'>}, 'nullable': False, 'example': [{'e': 'text', 't': 'Discussion'}], 'count': 60890, 'max_length': 1}
    link_flair_template_id: Mapped[Optional[str]] = mapped_column(Unicode(64))
    link_flair_text: Mapped[Optional[str]] = mapped_column(Unicode(128))
    link_flair_text_color: Mapped[Optional[str]] = mapped_column(Unicode(16))
    link_flair_type: Mapped[Optional[str]] = mapped_column(Unicode(8))
    locked: Mapped[Optional[bool]]
    media_metadata: Mapped[Optional[str]] = mapped_column(Unicode(4096))
    media_only: Mapped[Optional[bool]]
    name: Mapped[str] = mapped_column(String(16), index=True, unique=True)
    no_follow: Mapped[Optional[bool]]
    num_comments: Mapped[int]
    num_crossposts: Mapped[Optional[int]]
    num_reports: Mapped[Optional[int]]
    over_18: Mapped[bool]
    parent_whitelist_status: Mapped[Optional[str]] = mapped_column(Unicode(8))
    permalink: Mapped[str] = mapped_column(Unicode(128))
    pinned: Mapped[Optional[bool]]
    poll_data: Mapped[Optional[str]] = mapped_column(Unicode(4096))
    post_hint: Mapped[Optional[str]] = mapped_column(Unicode(16))
    previous_selftext: Mapped[Optional[str]] = mapped_column(Unicode(65536), deferred=True)
    quarantine: Mapped[Optional[bool]]
    removal_reason: Mapped[Optional[str]] = mapped_column(Unicode(8))
    removed_by_category: Mapped[Optional[str]] = mapped_column(Unicode(16))
    retrieved_on: Mapped[Optional[int]]
    retrieved_utc: Mapped[Optional[int]]
    rte_mode: Mapped[Optional[str]] = mapped_column(Unicode(8))
    score: Mapped[int]
    selftext: Mapped[str] = mapped_column(Unicode(65536), deferred=True)
    # selftext_html: Mapped[Optional[str]] = Unicode(4096)
    spoiler: Mapped[bool]
    stickied: Mapped[Optional[bool]]
    subreddit_name: Mapped[str] = mapped_column(ForeignKey("subreddit.name"), index=True)
    subreddit_id: Mapped[str] = mapped_column(String(12), index=True)
    subreddit_name_prefixed: Mapped[Optional[str]] = mapped_column(String(32))
    # subreddit_subscribers: Mapped[Optional[int]]
    # subreddit_type: Mapped[Optional[str]] = mapped_column(Unicode(8))
    thumbnail: Mapped[Optional[str]] = mapped_column(Unicode(256))
    thumbnail_height: Mapped[Optional[int]]
    thumbnail_width: Mapped[Optional[int]]
    title: Mapped[str] = mapped_column(Unicode(512), deferred=True)
    ups: Mapped[int]
    upvote_ratio: Mapped[int]
    url: Mapped[str] = mapped_column(Unicode(1024))
    url_overridden_by_dest: Mapped[Optional[str]] = mapped_column(Unicode(1024))

    author: Mapped["User"] = relationship(
        back_populates="posts",
        cascade="all",
        # primaryjoin="User.name == Post.author",
    )
    comments: Mapped[List["Comment"]] = relationship(
        back_populates="post",
        cascade="all,delete-orphan",
        # primaryjoin="User.name == Post.author",
    )
    root_comment: Mapped[Optional["Comment"]] = relationship(
        cascade="all",
        primaryjoin="Comment.id == foreign(Post.id)",
    )
    subreddit: Mapped["Subreddit"] = relationship(
        back_populates="posts",
        cascade="all",
    )

    def create_root_comment(self):
        """
        Create a placeholder root comment and return it.

        @return: the created root comment
        @rtype: L{Comment}
        """
        root_comment = Comment(
            author_name=ARCTICZIM_USERNAME,
            body="",
            controversiality=0,
            created_utc=self.created_utc,
            edited=0,
            gilded=0,
            id=self.id,
            parent_id=None,
            link_id=self.name,
            name=self.name,
            score=self.score,
            ups=0,
            permalink="",
            subreddit=self.subreddit,
        )
        # self.root_comment.post = self
        return root_comment

    @property
    def is_poll(self):
        """
        Check whether this post is a poll or not.

        @return: True if this post is a poll, False otherwise
        @rtype: L{bool}
        """
        return (self.poll_data is not None) and (self.poll_data != "false")

    @property
    def icon_name(self):
        """
        Return the name of the icon to use for this post type

        @return: the name of the icon to use
        @rtype: L{str}
        """
        if self.is_poll:
            return "poll"
        icons = {
            "rich:video": "video",
            "hosted:video": "video",
            "image": "img",
            "link": "link",
            "self": "text",
        }
        name = icons.get(self.post_hint, None)
        if name is None:
            if self.is_self:
                return "text"
            else:
                return "link"
        return name

    @property
    def is_crosspost(self):
        """
        Return True if this post is a crosspost

        @return: whether the post is a crosspost or not
        @rtype: L{str}
        """
        parsed = parse_reddit_url(self.url)
        if parsed is None:
            # not a reddit url at all
            return False
        elif parsed["type"] != "post":
            # not a post reference
            return False
        elif parsed["post"] == self.id:
            # URL simply points to the same post
            return False
        else:
            return True


class Comment(Base):
    """
    This model represents a comment in the database.
    """
    __tablename__ = "comment"
    __table_args__ = (
        UniqueConstraint("id"),
        UniqueConstraint("name"),
    )

    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    # all_awardings: {'always_present': False, 'types': {<class 'list'>}, 'nullable': False, 'example': [{'award_sub_type': 'GLOBAL', 'award_type': 'global', 'coin_price': 500, 'coin_reward': 100, 'count': 1, 'days_of_drip_extension': 0, 'days_of_premium': 7, 'description': 'Gives the author a week of Reddit Premium, %{coin_symbol}100 Coins to do with as they please, and shows a Gold Award.', 'end_date': None, 'giver_coin_reward': None, 'icon_format': None, 'icon_height': 512, 'icon_url': 'https://www.redditstatic.com/gold/awards/icon/gold_512.png', 'icon_width': 512, 'id': 'gid_2', 'is_enabled': True, 'is_new': False, 'name': 'Gold', 'penny_donate': None, 'penny_price': None, 'resized_icons': [{'height': 16, 'url': 'https://www.redditstatic.com/gold/awards/icon/gold_16.png', 'width': 16}, {'height': 32, 'url': 'https://www.redditstatic.com/gold/awards/icon/gold_32.png', 'width': 32}, {'height': 48, 'url': 'https://www.redditstatic.com/gold/awards/icon/gold_48.png', 'width': 48}, {'height': 64, 'url': 'https://www.redditstatic.com/gold/awards/icon/gold_64.png', 'width': 64}, {'height': 128, 'url': 'https://www.redditstatic.com/gold/awards/icon/gold_128.png', 'width': 128}], 'start_date': None, 'subreddit_coin_reward': 0, 'subreddit_id': None}], 'count': 605752, 'max_length': 6}
    archived: Mapped[Optional[bool]]
    author_name: Mapped[str] = mapped_column(ForeignKey("user.name"), index=True)
    author_cakeday: Mapped[Optional[bool]]
    author_created_utc: Mapped[Optional[int]]
    author_flair_background_color: Mapped[Optional[str]] = mapped_column(Unicode(8))
    author_flair_css_class: Mapped[Optional[str]] = mapped_column(Unicode(16))
    # author_flair_richtext: {'always_present': False, 'types': {<class 'list'>}, 'nullable': False, 'example': [{'e': 'text', 't': 'SYSTEM'}], 'count': 607055, 'max_length': 1}
    author_flair_template_id: Mapped[Optional[str]] = mapped_column(Unicode(64))
    author_flair_text: Mapped[Optional[str]] = mapped_column(Unicode(32))
    author_flair_text_color: Mapped[Optional[str]] = mapped_column(Unicode(4))
    author_flair_type: Mapped[Optional[str]] = mapped_column(Unicode(8))
    author_fullname: Mapped[Optional[str]] = mapped_column(Unicode(16))
    author_is_blocked: Mapped[Optional[bool]]
    author_patreon_flair: Mapped[Optional[bool]]
    author_premium: Mapped[Optional[bool]]
    body: Mapped[str] = mapped_column(Unicode(16384), deferred=True)
    body_html: Mapped[Optional[str]] = mapped_column(Unicode(4096), deferred=True)
    body_sha1: Mapped[Optional[str]] = mapped_column(Unicode(64), deferred=True)
    can_gild: Mapped[Optional[bool]]
    can_mod_post: Mapped[Optional[bool]]
    collapsed: Mapped[Optional[bool]]
    collapsed_reason: Mapped[Optional[str]] = mapped_column(Unicode(32))
    collapsed_reason_code: Mapped[Optional[str]] = mapped_column(Unicode(16))
    controversiality: Mapped[int]
    created_utc: Mapped[int]
    distinguished: Mapped[Optional[str]] = mapped_column(Unicode(16))
    downs: Mapped[Optional[int]]
    edited: Mapped[int]
    # expression_asset_data: {'always_present': False, 'types': {<class 'dict'>}, 'nullable': False, 'example': {'avatar_exp|145108495|bravo': {'avatar': {'e': 'Image', 'm': 'image/png', 's': {'u': 'https://i.redd.it/snoovatar/avatars/nftv2_bmZ0X2VpcDE1NToxMzdfZWI5NTlhNzE1ZGZmZmU2ZjgyZjQ2MDU1MzM5ODJjNDg1OWNiMTRmZV8xNDk4NDM2MQ_rare_e90add21-f3d6-4bce-8c20-105673bacfd7.png', 'x': 380, 'y': 600}}, 'expression': [{'e': ' Image', 'l': 'FRONT', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_1x.png', 'x': 150, 'y': 150}}, {'e': ' Image', 'l': 'BACK', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_1x_bg.png', 'x': 150, 'y': 150}}, {'e': ' Image', 'l': 'FRONT', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_2x.png', 'x': 300, 'y': 300}}, {'e': ' Image', 'l': 'BACK', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_2x_bg.png', 'x': 300, 'y': 300}}, {'e': ' Image', 'l': 'FRONT', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_3x.png', 'x': 500, 'y': 500}}, {'e': ' Image', 'l': 'BACK', 'm': 'image/png', 'n': 'bravo', 's': {'u': 'https://i.redd.it/snoovatar/expressions/bravo_1x1_3x_bg.png', 'x': 500, 'y': 500}}], 'perspective': 'CROPPED', 'position': 'CENTER', 'size': 'SIZE_1_X_1'}}, 'count': 5}
    gilded: Mapped[int]
    # gildings: {'always_present': False, 'types': {<class 'dict'>}, 'nullable': False, 'example': {'gid_1': 0, 'gid_2': 0, 'gid_3': 0}, 'count': 621211}
    id: Mapped[str] = mapped_column(Unicode(8), unique=True, index=True)
    is_submitter: Mapped[Optional[bool]]
    link_id: Mapped[str] = mapped_column(ForeignKey("post.name"), index=True)
    locked: Mapped[Optional[bool]]
    name: Mapped[str] = mapped_column(Unicode(16), unique=True, index=True)
    nest_level: Mapped[Optional[int]]
    no_follow: Mapped[Optional[bool]]
    num_reports: Mapped[Optional[int]]
    parent_id: Mapped[Optional[str]] = mapped_column(ForeignKey("comment.name"), index=True)
    permalink: Mapped[str] = mapped_column(Unicode(128))
    quarantined: Mapped[Optional[bool]]
    removal_reason: Mapped[Optional[str]] = mapped_column(Unicode(8))
    replies: Mapped[Optional[str]] = mapped_column(Unicode(128))
    # report_reasons: {'always_present': False, 'types': {<class 'list'>, <class 'NoneType'>}, 'nullable': True, 'example': None, 'count': 239684, 'max_length': 0}
    retrieved_on: Mapped[Optional[int]]
    rte_mode: Mapped[Optional[str]] = mapped_column(Unicode(8))
    saved: Mapped[Optional[bool]]
    score: Mapped[int]
    score_hidden: Mapped[Optional[bool]]
    send_replies: Mapped[Optional[bool]]
    stickied: Mapped[Optional[bool]]
    subreddit_name: Mapped[str] = mapped_column(ForeignKey("subreddit.name"))
    total_awards_received: Mapped[Optional[int]]
    ups: Mapped[int]

    author: Mapped["User"] = relationship(
        back_populates="comments",
        cascade="all",
        # primaryjoin="User.name == Post.author",
    )
    post: Mapped["Post"] = relationship(
        back_populates="comments",
        cascade="all",
        # primaryjoin="User.name == Post.author",
    )
    children: Mapped[List["Comment"]] = relationship(
        back_populates="parent",
        cascade="all,delete-orphan",
    )
    parent: Mapped["Comment"] = relationship(
        back_populates="children",
        cascade="all",
        remote_side=[name],
    )
    subreddit: Mapped["Subreddit"] = relationship(
        back_populates="comments",
        cascade="all",
    )

    @property
    def distinguished_class(self):
        """
        Return a string that can be used to identify if this comment was made by someone important.

        @return: a string if the commenter is somehow remarkable or L{None}
        @rtype: L{str} or L{None}
        """
        if self.distinguished:
            return self.distinguished
        elif self.is_submitter:
            return "submmitter"
        else:
            return None


class MediaFile(Base):
    """
    This model tracks the status of media files.

    This class keeps track of both URLs and checksums. The first mediafile
    that downloads a file is considered the primary one.
    """
    __tablename__ = "mediafile"
    __table_args__ = (
        UniqueConstraint("url"),
    )
    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    url: Mapped[str] = mapped_column(String(20248), unique=True, index=True)
    md5: Mapped[str] = mapped_column(String(32), index=True, nullable=True)
    mimetype: Mapped[str] = mapped_column(String(256), index=True, nullable=True)
    # extension: Mapped[str] = mapped_column(String(64))
    downloaded: Mapped[bool]
    size: Mapped[Optional[int]]
    primary_uid: Mapped[int] = mapped_column(ForeignKey("mediafile.uid"), nullable=True)

    primary: Mapped["MediaFile"] = relationship(
        back_populates="duplicates",
        cascade="all",
        remote_side=[uid],
    )
    duplicates: Mapped[List["MediaFile"]] = relationship(
        back_populates="primary",
        cascade="all,delete-orphan",
    )


class WikiPage(Base):
    """
    This model represents a subreddit wiki page.
    """
    __tablename__ = "wikipage"

    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    subreddit_name: Mapped[str] = mapped_column(ForeignKey("subreddit.name"), index=True)
    path: Mapped[str] = mapped_column(Unicode(512), deferred=True)
    content: Mapped[str] = mapped_column(Unicode(20 * 1024), deferred=True)
    revision_date: Mapped[int]
    revision_author: Mapped[str] = mapped_column(String(32))
    revision_reason: Mapped[Optional[str]] = mapped_column(Unicode(256), deferred=True)
    retrieved_on: Mapped[int]

    subreddit: Mapped["Subreddit"] = relationship(
        back_populates="wikipages",
    )

    @property
    def basepath(self):
        """
        Return the part of the path after the subreddit name.

        @return: the final part of the path
        @rtype: L{str}
        """
        path = self.path
        if path.startswith("/"):
            path = path[1:]
        if path.startswith("r/"):
            return path.split("/", 3)[-1]
        else:
            return self.path

    @property
    def title(self):
        """
        Placeholder for the title.

        @return: a title for the page
        @rtype: L{str}
        """
        return self.basepath


class SubredditRule(Base):
    """
    This class represents a subreddit rule.
    """
    __tablename__ = "subredditrule"

    uid: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    subreddit_name: Mapped[str] = mapped_column(ForeignKey("subreddit.name"))
    kind: Mapped[str] = mapped_column(String(16))
    priority: Mapped[int]
    short_name: Mapped[str] = mapped_column(Unicode(256), deferred=True)
    created_utc: Mapped[int]
    description: Mapped[str] = mapped_column(Unicode(1024), deferred=True)
    violation_reason: Mapped[str] = mapped_column(Unicode(512), deferred=True)

    subreddit: Mapped["Subreddit"] = relationship(
        back_populates="rules",
    )
